#
# Class to represent an email message (Message)
# and one to hold those in an Array (Messages)
#

require 'emaifilters'

class Messages < Array
  attr_accessor :by_uid, :first_to_show, :account
  def initialize(account, *args)
    super(*args)
    @account = account
    @by_uid = []
    @threads = nil
    @first_to_show = nil
  end

  def rec_add_threads(threads, parent, message, depth, pipes)
    subject = ""
    if message.seqno then
      m = self[message.seqno]
      if m then
        subject = m.envelope.subject
        if parent && self[parent.seqno] && \
          (self[parent.seqno].envelope.subject == subject || \
          self[parent.seqno].envelope.subject == subject.slice(4..-1))
        end
        if depth > 0
          if depth == 1 && parent == nil
            m.threaded_subject = "+-> "
          else
            space = " " * ([depth - 1, 0].max * 2) + "`"
            # threads below us need a connecting "|"
            pipes.each { |i|
              di = [i, 0].max * 2
              space[di] = "|" if space[di]
            }
            m.threaded_subject = space + "-> "
          end
        end
        m.threaded_index = threads.length
        threads << m
        if ((threads.first_to_show.nil? ||
             m.threaded_index < threads.first_to_show.threaded_index) &&
            (m.flags.index("O") || m.flags.index("N")))
          threads.first_to_show = m
        end
      end
      me = message
    elsif parent && self[parent.seqno]
      me = parent
    end
    p = pipes + [depth]
    message.children.each { |c|
      if c == message.children.last then p = pipes end
      rec_add_threads(threads, me, c, depth + 1, p)
    } if message
  end

  def threaded
    if @threads
      return @threads
    else
      @threads = Messages.new(self)
      threads_info = @account.get_threads_info
      threads_info.each { |msg| rec_add_threads(@threads, nil, msg, 0, []) }
      return @threads
    end
  end
end

class Header
  include EmaiFilters
  attr_accessor :raw, :decoded, :fields
  def initialize(header=nil)
    @raw    = header || ""
    @decoded = decode(@raw)
    @fields = header_to_fields(@decoded) || {}
  end
  
  def decode(header="")
    header.gsub!(/\r/,"")
    to = "UTF-8"
    transcode("ASCII", to, header)
  end

  def header_to_fields(h)
    return nil if h.nil?
    return {}  if h == ""
    fields = {}
    lines = h.split(/\r?\n(?=[A-Za-z][\w-]+:)\s*/) if h
    lines.each { |l|
      (k,v) = l.split(/:/,2)
      # transliterate content-type to Content-Type etc.
      k = k.split(/-/).collect {|h| h.capitalize}.join("-")
      fields[k] = v || ""
    } if lines
    fields
  end
end

class MIMEPart
  attr_accessor :header_fields, :children, :body, :decoded_body
  def initialize
    @header_fields = {} 
    @children      = []
    @decoded_body  = ""
    @body          = ""
  end

  def collect_body
    flat = ""
    bodies = @children.collect { |c| c.collect_body() }
    flat = bodies.join() if bodies

    @decoded_body + flat
  end
end

class Message
  include EmaiFilters
  attr_accessor :header, :info, :data, :threaded_subject, :envelope, :seqno, :threaded_index
  def initialize(account)
    @@debug           = $debug
    @account          = account 
    @body             = nil
    @header           = nil
    @flags            = ""
    @info             = ""
    @data             = {}
    @threaded_subject = ""
    @threaded_index   = nil
    @seqno            = nil
    @envelope         = nil
  end

  def uid
    @data["UID"]
  end

  def get_header
    @header ||= Header.new(@account.get_header(self))
  end

  def body
    @body ||= @account.get_body(self) or ""
  end

  def body=(body)
    @body = body
  end

  def decoded_body
    mpart = MIMEPart.new
    mpart.body = body() or "(empty message)"
    mpart.body.gsub!(/\r/,"")
    get_header()
    mpart.header_fields = @header.fields
    mpart.decoded_body = decode_mime(mpart)
    @decoded_body = mpart.collect_body() || "[MIME decoding failed. Showing raw body]\n#{mpart.body}"
  end
  alias_method :body_text, :decoded_body

  def body_formatted(width=76, tab_width=8)
    formatter = Text::Format.new {
      columns         = width
      left_margin     = 0
      right_margin    = 0
      first_indent    = 0 
      body_indent     = 0
      format_style    = Text::Format::LEFT_ALIGN
      extra_space     = false
      abbreviations   = {}
      tag_paragraph   = false
      tag_text        = []
      nobreak         = false
      nobreak_regex   = {}
      tabstop         = 8
      text            = nil
    }
    f_body = []
    my_body = decoded_body()
    tabs = Regexp.new("\t", nil, 'U')
    my_body.gsub!(tabs, " " * tab_width)
    my_body.split(/\r?\n/).each { |l|
      if l.length > width
        f_body += formatter.format(l).split("\n")
      else
        f_body << l
      end
    }
    f_body
  end

  # recursive
  # XXX this is a mess
  def decode_mime(main_part)
    my_body       = main_part.body
    header_fields = main_part.header_fields
    transfer_decoded_body = ""
    return_body   = ""
    trailing_part = ""
    return transfer_decoded_body if my_body.nil?

    case header_fields["Content-Transfer-Encoding"]
    when /quoted-printable/i
      transfer_decoded_body = my_body.unpack('M*')[0]
    when /base64/i
      transfer_decoded_body = my_body.unpack('m*')[0]
    else
      transfer_decoded_body = my_body
    end

    $logger.debug "before decode_mime:\n" +
                  "#{my_body}\n" +
                  "decode_mime: #{ header_fields.inspect }" if @@debug > 0

    if match = /attachment;.*?filename="([^"]+)"/im.match(header_fields["Content-Disposition"])
      return_body  << "\n----8<---- #{match[1]} ----8<----\n"
      trailing_part = "\n---->8---- #{match[1]} ---->8----\n"
    end

    if header_fields["Content-Type"].nil?
      $logger.debug "No Content-Type"  if @@debug > 0
      return_body += transfer_decoded_body.gsub(/\r/,"")
    elsif match = /^ ?((text|message)(\/(plain|html|x-diff|rfc822))?)( *;.*?charset="?([0-9A-Za-z\-]+)"?)?/i.match(header_fields["Content-Type"])
      $logger.debug "found text: #{match[1]} with charset #{match[6]}" if @@debug > 0

      if match[6] =~ /^iso-8859-[0-9]$/i then
        from = match[6]
        # XXX this should go into @config
        to = "UTF-8"
        transfer_decoded_body = transcode(from, to, transfer_decoded_body)
      end
      transfer_decoded_body.gsub!(/\r/,"")

      if match[1] && match[1] =~ /^text\/html$/i
        return_body << "----8<---- text/html ----8<----\n" +
                       "#{html2text(transfer_decoded_body)}" +
                       "\n---->8---- text/html ---->8----\n"
      else
        return_body << transfer_decoded_body
      end
    elsif match = /^ ?(application\/(pgp-signature))/im.match(header_fields["Content-Type"])
        return_body << transfer_decoded_body
    elsif match = /^ ?(multipart\/(alternative|signed|mixed|related|report)?) *;.*(micalg="pgp-([^"]+)";)?.*boundary="?([^"]+)"?/im.match(header_fields["Content-Type"])
      multipart = match[2]
      micalg    = match[3]
      boundary  = match[5]

      $logger.debug match.inspect if @@debug > 0
      $logger.debug "multipart: #{multipart}\nmatch: |#{match}|" if @@debug > 0

      parts = transfer_decoded_body.split(/--#{Regexp.escape(boundary)}(--)?\n/)

      $logger.debug "parts:",parts.inspect if @@debug > 0

      parts.each { |part|
        $logger.debug "part\n |#{part}|" if @@debug > 0
        unless part =~ /\S/m
          $logger.debug "skipping" if @@debug > 0
          next
        end

        mpart = MIMEPart.new
        (part_header, part_body) = part.split(/\n\n/, 2)
        mpart.header_fields = Header.new(part_header).fields
        mpart.body = part_body

        if match = /^ ?(application\/(pgp-signature))/im.match(mpart.header_fields["Content-Type"])
          # Make it look like inlined signature so $EDITOR can --verify
          mpart.decoded_body = mpart.body
          mpart_h = MIMEPart.new
          hash = ""
          hash = "Hash: #{micalg}\n" if micalg
          mpart_h.body = mpart_h.decoded_body = "-----BEGIN PGP SIGNED MESSAGE-----\n#{hash}\n"
          m = main_part.children.pop
          # This takes too much time. Leave it to $EDITOR
          #mpart.decoded_body << pgp_verify(m.decoded_body, mpart.body)
          main_part.children << mpart_h << m
        else
          # recursion
          mpart.decoded_body = decode_mime(mpart)
        end

        main_part.children << mpart

      } if parts

    else # Unknown Type
        return_body << "----------\n" +
                       "| Name: #{header_fields['name']}\n" +
                       "| Filename: #{header_fields['filename']}\n" +
                       "| Content-Type: #{header_fields['Content-Type']}\n" +
                       "----8<----\n" +
                       transfer_decoded_body +
                       "----8<----\n"
      $logger.warn "unknown Content-Type: <#{header_fields['Content-Type']}>"
    end
    return_body + trailing_part
  end

  def decoded_subject
    decoded_subject = envelope.subject || '(no subject)'
    while match = decoded_subject.match(/=\?([A-Za-z0-9\-]+)\?(.)\?(.+)\?=/)
      char_set = match[1]
      encoding = match[2]
      string   = match[3]
      if encoding =~ /^B$/i then
        string = string.unpack('m*')[0]
      end
      string = string.unpack('M*')[0]
      decoded_subject.gsub!(/#{Regexp.escape(match[0])}/, string)
    end
    decoded_subject.gsub!(/\t/,"    ")
    decoded_subject
  end

  def size_short
    in_k = size / 1024
    if in_k > 0
      in_m = in_k / 1024
      if in_m > 0
        in_g = in_m / 1024
        if in_g > 0
          s = ">=1G"
        else
          s = in_m.modulo(1024).to_s + "M"
        end
      else
        s = in_k.modulo(1024).to_s + "K"
      end
    else
      s = size.to_s + "B"
    end
    s
  end

  def flags
    seen = false
    new  = false
    flags = @data["FLAGS"].collect { |f|
      case f
      when :Seen
        seen = true
        next
      when :Recent
        new = true
        "N"
      when :Answered
        answered = true
        "A"
      when :Deleted
        deleted = true
        "D"
      when :Flagged
        flagged = true
        "F"
      when :Draft
        draft = true
        "d"
      else
        $logger.warn "UKNOWN FLAG: #{f}"
        "U"
      end
    }
    # old
    if not (seen or new) then flags << "O" end
    flags
  end

  def size
    @data["RFC822.SIZE"]
  end
end

# vim: ts=2 sw=2
